"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _react = _interopRequireWildcard(require("react"));

var normalizeValue = function normalizeValue(min, value, max) {
  return Math.max(Math.min(value, max), min);
};

var MIN_INDEX = 0;

var ViewPortList = _react.default.forwardRef(function (_ref, ref) {
  var _ref$viewPortRef = _ref.viewPortRef,
      viewPortRef = _ref$viewPortRef === void 0 ? null : _ref$viewPortRef,
      _ref$items = _ref.items,
      items = _ref$items === void 0 ? [] : _ref$items,
      itemMinHeight = _ref.itemMinHeight,
      _ref$marginBottom = _ref.marginBottom,
      marginBottom = _ref$marginBottom === void 0 ? 0 : _ref$marginBottom,
      _ref$overscan = _ref.overscan,
      overscan = _ref$overscan === void 0 ? 1 : _ref$overscan,
      _ref$initialIndex = _ref.initialIndex,
      initialIndex = _ref$initialIndex === void 0 ? 0 : _ref$initialIndex,
      _ref$initialAlignToTo = _ref.initialAlignToTop,
      initialAlignToTop = _ref$initialAlignToTo === void 0 ? true : _ref$initialAlignToTo,
      children = _ref.children;
  var maxIndex = items.length - 1;
  var normalizedInitialIndex = normalizeValue(MIN_INDEX, initialIndex, maxIndex);

  var _useState = (0, _react.useState)([normalizedInitialIndex, normalizedInitialIndex]),
      _useState2 = (0, _slicedToArray2.default)(_useState, 2),
      _useState2$ = (0, _slicedToArray2.default)(_useState2[0], 2),
      startIndex = _useState2$[0],
      endIndex = _useState2$[1],
      setIndexes = _useState2[1];

  var topRef = (0, _react.useRef)(null);
  var bottomRef = (0, _react.useRef)(null);
  var cache = (0, _react.useRef)([]);
  var step = (0, _react.useRef)(function () {});

  var _scrollToIndex = (0, _react.useRef)(normalizedInitialIndex ? {
    index: normalizedInitialIndex,
    alignToTop: initialAlignToTop
  } : null);

  var isOverflowAnchorSupported = (0, _react.useRef)(CSS.supports('overflow-anchor: auto'));
  var scrollCompensationEndIndex = (0, _react.useRef)(null);
  var topStyle = (0, _react.useMemo)(function () {
    var topHeight = cache.current.slice(MIN_INDEX, startIndex).reduce(function (sum, next) {
      return sum + next;
    }, startIndex * (itemMinHeight + marginBottom));
    return {
      minHeight: topHeight,
      height: topHeight,
      maxHeight: topHeight,
      overflowAnchor: 'none',
      pointerEvents: 'none',
      userSelect: 'none',
      padding: 0,
      margin: 0
    };
  }, [startIndex, itemMinHeight, marginBottom]);
  var bottomStyle = (0, _react.useMemo)(function () {
    var bottomHeight = cache.current.slice(endIndex + 1, maxIndex).reduce(function (sum, next) {
      return sum + next;
    }, (itemMinHeight + marginBottom) * (maxIndex - endIndex));
    return {
      minHeight: bottomHeight,
      height: bottomHeight,
      maxHeight: bottomHeight,
      overflowAnchor: 'none',
      pointerEvents: 'none',
      userSelect: 'none',
      padding: 0,
      margin: 0
    };
  }, [endIndex, maxIndex, itemMinHeight, marginBottom]);
  var slicedItems = (0, _react.useMemo)(function () {
    return items.slice(startIndex, endIndex + 1).map(function (item, index) {
      return children(item, startIndex + index);
    });
  }, [items, startIndex, endIndex, children]);

  step.current = function () {
    var firstElement = topRef.current.nextSibling;

    if (_scrollToIndex.current) {
      var targetIndex = normalizeValue(MIN_INDEX, _scrollToIndex.current.index, maxIndex);

      if (startIndex !== targetIndex) {
        setIndexes([targetIndex, targetIndex]);
        return;
      }

      firstElement.scrollIntoView(_scrollToIndex.current.alignToTop);
      _scrollToIndex.current = null;
      return;
    }

    var clientHeight = document.documentElement.clientHeight;
    var viewPortRect = viewPortRef && viewPortRef.current && viewPortRef.current.getBoundingClientRect();
    var viewPortTop = viewPortRect ? normalizeValue(MIN_INDEX, viewPortRect.top, clientHeight) : MIN_INDEX;
    var viewPortBottom = viewPortRect ? normalizeValue(MIN_INDEX, viewPortRect.bottom, clientHeight) : clientHeight;
    var overscanHeight = overscan * (itemMinHeight + marginBottom);
    var topLimit = viewPortTop - overscanHeight;
    var bottomLimit = viewPortBottom + overscanHeight;
    var firstElementRect = firstElement.getBoundingClientRect();
    var lastElementRect = bottomRef.current.previousSibling.getBoundingClientRect();
    var nextStartIndex = startIndex;
    var nextEndIndex = endIndex;

    if (firstElementRect.top >= bottomLimit) {
      var diff = firstElementRect.top - bottomLimit;
      --nextStartIndex;

      while (diff >= 0 && nextStartIndex >= MIN_INDEX) {
        diff -= cache.current[nextStartIndex] || itemMinHeight;
        --nextStartIndex;
      }

      scrollCompensationEndIndex.current = startIndex;
      nextEndIndex = nextStartIndex;
    } else if (lastElementRect.bottom + marginBottom <= topLimit) {
      var _diff = topLimit - lastElementRect.bottom + marginBottom;

      nextStartIndex = endIndex + 1;

      while (_diff >= 0 && nextStartIndex <= maxIndex) {
        _diff -= cache.current[nextStartIndex] || itemMinHeight;
        ++nextStartIndex;
      }

      nextEndIndex = nextStartIndex;
    } else {
      if (firstElementRect.bottom + marginBottom < topLimit) {
        ++nextStartIndex;
      } else if (firstElementRect.top >= topLimit) {
        scrollCompensationEndIndex.current = startIndex;
        --nextStartIndex;
      }

      if (lastElementRect.bottom + marginBottom <= bottomLimit) {
        ++nextEndIndex;
      } else if (lastElementRect.top > bottomLimit) {
        --nextEndIndex;
      }
    }

    nextStartIndex = normalizeValue(MIN_INDEX, nextStartIndex, maxIndex);
    nextEndIndex = normalizeValue(nextStartIndex, nextEndIndex, maxIndex);

    if (nextStartIndex !== startIndex || nextEndIndex !== endIndex) {
      var index = startIndex;
      var element = firstElement;

      while (index <= endIndex && element !== bottomRef.current) {
        cache.current[index] = element.clientHeight - itemMinHeight;
        element = element.nextSibling;
        ++index;
      }

      setIndexes([nextStartIndex, nextEndIndex]);
    }
  };

  (0, _react.useImperativeHandle)(ref, function () {
    return {
      scrollToIndex: function scrollToIndex() {
        var index = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : -1;
        var alignToTop = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
        _scrollToIndex.current = {
          index: index,
          alignToTop: alignToTop
        };
      }
    };
  }, []);
  (0, _react.useLayoutEffect)(function () {
    if (scrollCompensationEndIndex.current === null) {
      return;
    }

    if (!viewPortRef || isOverflowAnchorSupported.current) {
      scrollCompensationEndIndex.current = null;
      return;
    }

    var index = startIndex;
    var element = topRef.current.nextSibling;
    var heightDiff = 0;

    while (index < scrollCompensationEndIndex.current && element !== bottomRef.current) {
      heightDiff += element.clientHeight - (cache.current[index] || 0) - itemMinHeight;
      element = element.nextSibling;
      ++index;
    }

    if (viewPortRef && heightDiff) {
      viewPortRef.current.style.overflowY = 'hidden';
      viewPortRef.current.scrollTop += heightDiff;
      viewPortRef.current.style.overflowY = null;
    }

    scrollCompensationEndIndex.current = null;
  }, [viewPortRef, startIndex, itemMinHeight]);
  (0, _react.useEffect)(function () {
    var frameId = MIN_INDEX;

    var frame = function frame() {
      frameId = requestAnimationFrame(frame);
      step.current();
    };

    frame();
    return function () {
      return cancelAnimationFrame(frameId);
    };
  }, []);
  return /*#__PURE__*/_react.default.createElement(_react.Fragment, {
    key: "ViewPortList"
  }, /*#__PURE__*/_react.default.createElement("div", {
    key: "ViewPortListTop",
    ref: topRef,
    style: topStyle
  }), slicedItems, /*#__PURE__*/_react.default.createElement("div", {
    key: "ViewPortListBottom",
    ref: bottomRef,
    style: bottomStyle
  }));
});

var _default = ViewPortList;
exports.default = _default;